from pathlib import Path
import numpy as np
import re
import itertools
from typing import Generator, List, Optional, Sequence, Set, Tuple, Dict, Iterator, Any
import xarray
import pandas as pd
from collections.abc import MutableMapping
from itertools import product, combinations, chain
from glob import glob
from collections import defaultdict
from datetime import datetime
from parse import extract_format, compile
from functools import cmp_to_key, lru_cache
import pandas as pd
import os
import string


def is_singular(value):
    """Whether a value is singular or has multiple options.
    """
    if isinstance(value, str):
        return True
    try:
        iter(value)
        return False
    except TypeError:
        return True

class Placeholders(MutableMapping):
    """Dictionary-like object containing the placeholder values.

    It understands about sub-trees (i.e., if "<sub_tree>/<placeholder>" does not exist it will return "<placeholder>" instead).
    """
    def __init__(self, *args, **kwargs):
        self.mapping = {}
        self.linkages = {}
        self.update(dict(*args, **kwargs))

    def copy(self):
        p = Placeholders()
        p.mapping = dict(self.mapping)
        p.linkages = dict(self.linkages)
        return p

    def __getitem__(self, key: str):
        actual_key = self.find_key(key)
        if actual_key is None:
            raise KeyError(f"No parameter value available for {key}")
        if actual_key in self.linkages:
            return self.mapping[self.linkages[actual_key]][actual_key]
        return self.mapping[actual_key]

    def __delitem__(self, key):
        if isinstance(key, tuple):
            key = frozenset(key)
        del self.mapping[key]
        if isinstance(key, frozenset):
            for k in key:
                del self.linkages[k]

    def __setitem__(self, key, value):
        if isinstance(key, tuple):  # create linked placeholders
            if len(key) != len(value):
                raise ValueError(f"Attempting to set linked placeholders for {key}, but {value} has a different number of elements than {key}")
            if any([len(value[0]) != len(v) for v in value]):
                raise ValueError(f"Attempting to set linked placeholders for {key}, but not all elements in {value} have the same length")
            value = {k: v for k, v in zip(key, value)}
            key = frozenset(key)
        if isinstance(key, frozenset):
            for k in list(key):
                if k in self.linkages:
                    unmatched_keys = [unmatched for unmatched in self.linkages[k] if unmatched not in key]
                    if len(unmatched_keys) > 0:
                        raise ValueError(f"Attempting to set linked placeholders for {key}, but {k} is already linked to {unmatched_keys}")
            self.mapping[key] = value
            for k in list(key):
                self.linkages[k] = key
                if k in self.mapping:
                    del self.mapping[k]
        elif key in self.linkages:
            old_values = self.mapping[self.linkages[key]]
            if is_singular(value):
                if value in old_values[key]:
                    self.unlink(*old_values.keys())
                    idx = old_values[key].index(value)
                    for skey in old_values:
                        self.mapping[skey] = old_values[skey][idx]
                else:
                    raise ValueError(f"Can not overwrite placeholder {key} with new values as it is linked to: {self.linkages[key]}")
            else:
                if all(v in old_values[key] for v in value):
                    idx = [old_values[key].index(v) for v in value]
                    for skey in old_values:
                        old_values[skey] = tuple(old_values[skey][i] for i in idx)
                else:
                    raise ValueError(f"Can not overwrite placeholder {key} with new values as it is linked to: {self.linkages[key]}")
        else:
            self.mapping[key] = value

    def __iter__(self):
        for key in self.mapping:
            if self.mapping[key] is not None:
                yield key

    def __len__(self):
        return len(self.mapping)

    def __repr__(self):
        return f"Placeholders({self.mapping})"

    def find_key(self, key: str) -> Optional[str]:
        """Finds the actual key containing the value

        Will look for:

            - not None value for the key itself
            - not None value for any parent (i.e, for key "A/B", will look for "B" as well)
            - otherwise will return None

        Args:
            key (str): placeholder name

        Returns:
            Optional[str]: None if no value for the key is available, otherwise the key used to index the value
        """
        if not isinstance(key, str):
            key = frozenset(key)
        elif key in self.linkages:
            return key
        if self.mapping.get(key, None) is not None:
            return key
        elif '/' in key:
            *sub_trees, _, actual_key = key.split('/')
            new_key = '/'.join([*sub_trees, actual_key])
            return self.find_key(new_key)
        else:
            return None

    def split(self) -> Tuple["Placeholders", "Placeholders"]:
        """Splits all placeholders into those with a single value or those with multiple values

        Placeholders are considered to have multiple values if they are equivalent to 1D-arrays (lists, tuples, 1D ndarray, etc.). 
        Anything else is considered a single value (string, int, float, etc.)

        Args:
            placeholders (Dict): all mappings from placeholder names to values

        Returns:
            Tuple[Dict, Dict]: Returns tuples with two dictionaries (first those with single values, then those with the multiple values)
        """
        single_placeholders = Placeholders()
        multi_placeholders = Placeholders()
        for name, value in self.mapping.items():
            if isinstance(name, frozenset) or not is_singular(value):
                multi_placeholders[name] = value
            else:
                single_placeholders[name] = value
        return single_placeholders, multi_placeholders

    def iter_over(self, keys) -> Generator["Placeholders", None, None]:
        """Iterate over the placeholder placeholder names

        Args:
            keys (Sequence[str]): sequence of placeholder names to iterate over

        Returns:
            Generator[FileTree]: yields Placeholders object, where each of the listed keys only has a single possible value
        """
        actual_keys = [self.linkages.get(self.find_key(key), key) for key in keys]
        unfilled = {orig for orig, key in zip(keys, actual_keys) if key is None}
        if len(unfilled) > 0:
            raise KeyError(f"Can not iterate over undefined placeholders: {unfilled}")

        unique_keys = []
        iter_values = {}
        for key in actual_keys:
            if key not in unique_keys:
                if isinstance(key, frozenset):  # linked placeholder
                    unique_keys.append(key)
                    iter_values[key] = [{k: self[k][idx] for k in key} for idx in range(len(self[list(key)[0]]))]
                elif not is_singular(self[key]):  # iterable placeholder
                    unique_keys.append(key)
                    iter_values[key] = self[key]

        for values in product(*[iter_values[k] for k in unique_keys]):
            new_vars = Placeholders(self)
            for key, value in zip(unique_keys, values):
                if isinstance(key, frozenset):
                    del new_vars[key]  # break the placeholders link
                    new_vars.update(value)
                else:
                    new_vars[key] = value
            yield new_vars

    def link(self, *keys):
        """
        Link the placeholders represented by `keys`.

        When iterating over linked placeholders the i-th tree will contain the i-th element from all linked placeholders,
        instead of the tree containing all possible combinations of placeholder values.

        This can be thought of using `zip` for linked variables and `itertools.product` for unlinked ones.
        """
        self[keys] = [self[key] for key in keys]

    def unlink(self, *keys):
        """
        Unlink the placeholders represented by `keys`.

        See :meth:`link` for how linking affects the iteration through placeholders with multiple values.

        Raises a ValueError if the placeholders are not actually linked.
        """
        if keys not in self:
            raise ValueError(f"{keys} were not linked, so cannot unlink them")
        new_vars = {k: self[k] for k in keys}
        del self[keys]
        self.update(new_vars)

class MyDataArray:
    """Wrapper around xarray.DataArray for internal usage

    It tries to delay creating the DataArray object as long as possible (as using them for small arrays is slow...)
    """
    def __init__(self, data, coords=None):
        self.as_xarray = coords is None
        if self.as_xarray:
            assert isinstance(data, xarray.DataArray)
            self.data_array = data
        else:
            self.data = data
            self.coords = coords

    def map(self, func) -> "MyDataArray":
        if self.as_xarray:
            return MyDataArray(xarray.apply_ufunc(func, self.data_array, vectorize=True))
        else:
            return MyDataArray(np.array([func(d) for d in self.data.flat]).reshape(self.data.shape), self.coords)

    def to_xarray(self, ) -> xarray.DataArray:
        if self.as_xarray:
            return self.data_array
        else:
            return xarray.DataArray(self.data, [_to_index(name, values) for name, values in self.coords])

    @staticmethod
    def concat(parts, new_index) -> "MyDataArray":
        if len(parts) == 0:
            return MyDataArray(np.array([]), [])
        to_xarray = (
            any(p.as_xarray for p in parts) or
            any(
                len(p.coords) != len(parts[0].coords) or
                any(np.all(name1 != name2) for (name1, _), (name2, _) in zip(p.coords, parts[0].coords))
                for p in parts
            )
        )
        if to_xarray:
            return MyDataArray(xarray.concat([p.to_xarray() for p in parts], _to_index(*new_index)))
        else:
            new_data = np.stack([p.data for p in parts], axis=0)
            new_coords = list(parts[0].coords)
            new_coords.insert(0, new_index)
            return MyDataArray(new_data, new_coords)


def _to_index(name, values):
    if isinstance(name, str):
        return pd.Index(values, name=name)
    else:
        return pd.MultiIndex.from_tuples(values, names=name)



class Template:
    def __init__(self, parent: "Template", unique_part: str):
        self.parent = parent
        self.unique_part = unique_part

    @property
    def as_path(self, ) -> Path:
        """The full path with no placeholders filled in
        """
        if self.parent is None:
            return Path(self.unique_part)
        return self.parent.as_path.joinpath(self.unique_part)

    @property
    def as_string(self, ):
        if self.parent is None:
            return str(self.unique_part)
        return os.path.join(self.parent.as_string, str(self.unique_part))

    def __str__(self, ):
        return f"Template({self.as_string})"

    def children(self, templates: Sequence["Template"]) -> List["Template"]:
        """From a sequence of templates find the children

        Returns:
            List[Template]: list of children templates
        """
        return [t for t in templates if t.parent is self]

    def as_multi_line(self, other_templates: Dict[str, "Template"], indentation=4) -> str:
        """Generates a string describing this and child templates as 

        Args:
            other_templates (Dict[str, Template]): templates including all the child templates and itself
            indentation (int, optional): number of spaces to use as indentation. Defaults to 4

        Returns:
            str: multi-line string that can be processed by :meth:`file_tree.FileTree.read`
        """
        result = self._as_multi_line_helper(other_templates, indentation)

        is_top_level = other_templates[""] is self
        if not is_top_level and self.parent is None:
            return '!' + result
        else:
            return result

    def _as_multi_line_helper(self, other_templates: Dict[str, "Template"], indentation=4, _current_indentation=0) -> str:
        leaves = []
        branches = []
        for t in sorted(self.children(other_templates.values()), key=lambda t: t.unique_part):
            if len(t.children(other_templates.values())) == 0:
                leaves.append(t)
            else:
                branches.append(t)

        is_top_level = other_templates[""] is self
        if is_top_level:
            base_line = '.'
            assert _current_indentation == 0 and self.parent is None
            _current_indentation = -indentation
        else:
            base_line = _current_indentation * ' ' + self.unique_part

        all_keys = []
        for key, value in other_templates.items():
            if value is not self:
                continue
            if is_top_level and key == "":
                continue
            all_keys.append(key)
        if is_top_level and len(all_keys) == 0:
            lines = []
        elif len(all_keys) == 1 and all_keys[0] == self.guess_key():
            lines = [base_line]
        else:
            assert len(all_keys) > 0
            lines = [base_line + f' ({",".join(all_keys)})']

        already_done = set()
        for t in leaves + branches:
            if t not in already_done:
                lines.append(t._as_multi_line_helper(other_templates, indentation, indentation + _current_indentation))
                already_done.add(t)
        return '\n'.join(lines)

    @property
    def _parts(self, ):
        return TemplateParts.parse(self.as_string)

    def placeholders(self, valid=None) -> List[str]:
        """Returns a list of the placeholder names

        Returns:
            List[str]: placeholder names in order that they appear in the template
        """
        return self._parts.ordered_placeholders(valid)

    def format_single(self, placeholders: Placeholders, check=True, keep_optionals=False) -> str:
        """Formats the template with the placeholders filled in

        Only placeholders with a single value are considered.

        Args:
            placeholders (Placeholders): values to fill into the placeholder
            check (bool): skip check for missing placeholders if set to True
            keep_optionals: if True keep optional parameters that have not been set (will cause the check to fail)

        Raises:
            KeyError: if any placeholder is missing

        Returns:
            str: filled in template
        """
        single_placeholders, _ = placeholders.split()
        template = self._parts.fill_single_placeholders(single_placeholders)
        if not keep_optionals:
            template = template.remove_optionals()
        if check:
            unfilled = template.required_placeholders()
            if len(unfilled) > 0:
                raise KeyError(f"Missing placeholder values for {unfilled}")
        return str(template)

    def format_mult(self, placeholders: Placeholders, check=False, filter=False, matches=None) -> xarray.DataArray:
        """Replaces placeholders in template with the provided placeholder values

        Args:
            placeholders (Placeholders): mapping from placeholder names to single or multiple vaalues
            check (bool): skip check for missing placeholders if set to True
            filter (bool): filter out non-existing files if set to True

        Raises:
            KeyError: if any placeholder is missing

        Returns:
            xarray.DataArray: array with possible resolved paths.
                If `filter` is set to True the non-existent paths are replaced by None
        """
        parts = self._parts
        resolved = parts.resolve(placeholders)
        if check:
            for template in resolved.data.flatten():
                unfilled = template.required_placeholders()
                if len(unfilled) > 0:
                    raise KeyError(f"Missing placeholder values for {unfilled}")
        paths = resolved.map(lambda t: str(t))
        if not filter:
            return paths.to_xarray()
        placeholder_dict = dict(placeholders)
        path_matches = [str(parts.fill_single_placeholders(Placeholders({**placeholder_dict, **match})).remove_optionals())
                        for match in (self.all_matches(placeholders) if matches is None else matches)]
        return paths.map(lambda p: p if p in path_matches else '').to_xarray()

    def optional_placeholders(self, ) -> Set[str]:
        """Finds all placeholders that are only within optional blocks (i.e., they do not require a value)

        Returns:
            Set[str]: names of optional placeholders
        """
        return self._parts.optional_placeholders()

    def required_placeholders(self, ) -> Set[str]:
        """Finds all placeholders that are outside of optional blocks (i.e., they do require a value)

        Returns:
            Set[str]: names of required placeholders
        """
        return self._parts.required_placeholders()

    def guess_key(self, ) -> str:
        """Proposes a short name for the template

        The proposed short name is created by:

            - taking the basename (i.e., last component) of the path
            - removing the first '.' and everything beyond (to remove the extension)

        .. warning::

            If there are multiple dots within the path's basename, this might remove far more than just the extension.

        Returns:
            str: proposed short name for this template (used if user does not provide one)
        """
        parts = self.as_path.parts
        if len(parts) == 0:
            return ""
        else:
            return parts[-1].split('.')[0]

    def add_precursor(self, text) -> "Template":
        """Returns a new Template with any placeholder names in the unique part now preceded by `text`

        Used for adding sub-trees
        """
        parts = TemplateParts.parse(self.unique_part).parts
        updated = ''.join([str(p.add_precursor(text)) for p in parts])
        return Template(self.parent, updated)

    def get_all_placeholders(self, placeholders: Placeholders, matches=None) -> Placeholders:
        """Fill placeholders with possible values based on what is available on disk

        Args:
            placeholders (Placeholders): New values for undefined placeholders in template
        """
        undefined = defaultdict(set)
        for match in self.all_matches(placeholders) if matches is None else matches:
            for name, value in match.items():
                if name not in placeholders:
                    undefined[name].add(value)

        def cmp(item1, item2):
            if item1 is None:
                return -1
            if item2 is None:
                return 1
            if item1 < item2:
                return -1
            if item1 > item2:
                return 1
            return 0

        return Placeholders({k: sorted(v, key=cmp_to_key(cmp)) for k, v in undefined.items()})

    def all_matches(self, placeholders: Placeholders):
        """Returns a sequence of all possible variable values matching existing files on disk

        Only variable values matching existing placeholder values are returned (undefined placeholders are unconstrained).
        """
        single_vars, multi_vars = placeholders.split()
        filled_template: TemplateParts = self._parts.fill_single_placeholders(single_vars)
        res = []
        for match in filled_template.all_matches():

            def check_name_with_edit(name):
                value = match[name]
                if name in multi_vars and multi_vars.find_key(name) == name:
                    if value not in multi_vars[name]:
                        return False
                    return True
                if '/' in name:
                    del match[name]
                    parent_name = name[name.find('/') + 1:]
                    if parent_name in match:
                        return match[parent_name] == value
                    match[parent_name] = value
                    return check_name_with_edit(parent_name)
                return True

            if not all(check_name_with_edit(name) for name in list(match.keys())):
                continue
            res.append(match)
        return res


def extract_placeholders(template, filename, known_vars=None):
    """
    Extracts the placeholder values from the filename

    :param template: template matching the given filename
    :param filename: filename
    :param known_vars: already known placeholders
    :return: dictionary from placeholder names to string representations (unused placeholders set to None)
    """
    return TemplateParts.parse(template).extract_placeholders(filename, known_vars)



class Part:
    """
    Individual part of a template

    3 subclasses are defined:

    - :class:`Literal`: piece of text
    - :class:`Required`: required placeholder to fill in (between curly brackets)
    - :class:`OptionalPart`: part of text containing optional placeholders (between square brackets)
    """
    def fill_single_placeholders(self, placeholders: Placeholders, ignore_type=False) -> Sequence["Part"]:
        """
        Fills in the given placeholders
        """
        return (self, )

    def optional_placeholders(self, ) -> Set["Part"]:
        """
        Returns all placeholders in optional parts
        """
        return set()

    def required_placeholders(self, ) -> Set["Part"]:
        """
        Returns all required placeholders
        """
        return set()

    def contains_optionals(self, placeholders: Set["Part"]=None):
        """
        Returns True if this part contains the optional placeholders
        """
        return False

    def append_placeholders(self, placeholders: List[str], valid=None):
        """
        Appends the placeholders in this part to the provided list in order
        """
        pass

    def add_precursor(self, text: str) -> "Part":
        """Prepends any placeholder names by `text`.
        """
        return self

    def for_defined(self, placeholder_names: Set[str]) -> "Part":
        """Returns the template string assuming the placeholders in `placeholder_names` are defined

        Removes any optional parts, whose placeholders are not in `placeholder_names`.
        """
        return [self]

    def remove_precursors(self, placeholders=None):
        return self

class Literal(Part):
    def __init__(self, text: str):
        """
        Literal part is defined purely by the text it contains

        :param text: part of the template
        """
        self.text = text

    def __str__(self):
        """
        Returns this part of the template as a string
        """
        return self.text

    def __eq__(self, other):
        if not isinstance(other, Literal):
            return NotImplemented
        return self.text == other.text


class Required(Part):
    def __init__(self, var_name, var_formatting=None):
        """
        Required part of template (between curly brackets)

        Required placeholder part of template is defined by placeholder name and its format

        :param var_name: name of placeholder
        :param var_formatting: how to format the placeholder
        """
        self.var_name = var_name
        self.var_formatting = var_formatting

    def __str__(self):
        """
        Returns this part of the template as a string
        """
        if self.var_formatting is None or len(self.var_formatting) == 0:
            return '{' + self.var_name + '}'
        else:
            return '{' + self.var_name + ':' + self.var_formatting + '}'

    def fill_single_placeholders(self, placeholders: Placeholders, ignore_type=False):
        value = placeholders.get(self.var_name, None)
        if value is None:
            return (self, )
        else:
            if not ignore_type and len(self.var_formatting) > 0:
                format_type = extract_format(self.var_formatting, [])["type"]
                if format_type in list(r"dnbox"):
                    value = int(value)
                elif format_type in list(r"f%eg"):
                    value = float(value)
                elif format_type in ['t' + ft for ft in 'iegachs']:
                    value = datetime(value)
            res = TemplateParts.parse(format(value, '' if ignore_type else self.var_formatting))
            if len(res.parts) == 1:
                return res.parts
            return res.fill_single_placeholders(placeholders, ignore_type=ignore_type).parts

    def required_placeholders(self, ):
        return {self.var_name}

    def append_placeholders(self, placeholders, valid=None):
        if valid is not None and self.var_name not in valid:
            raise ValueError(f"Placholder {self.var_name} is not defined")
        placeholders.append(self.var_name)

    def add_precursor(self, text: str) -> "Required":
        """Prepends any placeholder names by `text`.
        """
        return Required(text + self.var_name, self.var_formatting)

    def remove_precursors(self, placeholders=None):
        if placeholders is None:
            new_name = self.var_name.split('/')[-1]
        else:
            key =  placeholders.find_key(self.var_name)
            new_name = self.var_name if key is None else key
        return Required(new_name, self.var_formatting)

    def __eq__(self, other):
        if not isinstance(other, Required):
            return NotImplemented
        return (self.var_name == other.var_name) & (self.var_formatting == other.var_formatting)

class OptionalPart(Part):
    def __init__(self, sub_template: "TemplateParts"):
        """
        Optional part of template (between square brackets)

        Optional part can contain literal and required parts

        :param sub_template: part of the template within square brackets
        """
        self.sub_template = sub_template

    def __str__(self):
        return '[' + str(self.sub_template) + ']'

    def fill_single_placeholders(self, placeholders: Placeholders, ignore_type=False):
        new_opt = self.sub_template.fill_single_placeholders(placeholders, ignore_type=ignore_type)
        if len(new_opt.required_placeholders()) == 0:
            return (Literal(str(new_opt)), )
        return (OptionalPart(new_opt), )

    def optional_placeholders(self, ):
        return self.sub_template.required_placeholders()

    def contains_optionals(self, placeholders=None):
        if placeholders is None and len(self.optional_placeholders()) > 0:
            return True
        return len(self.optional_placeholders().intersection(placeholders)) > 0

    def append_placeholders(self, placeholders, valid=None):
        try:
            placeholders.extend(self.sub_template.ordered_placeholders(valid=valid))
        except ValueError:
            pass

    def add_precursor(self, text: str) -> "OptionalPart":
        return OptionalPart(TemplateParts([p.add_precursor(text) for p in self.sub_template.parts]))

    def for_defined(self, placeholder_names: Set[str]) -> "Part":
        """Returns the template string assuming the placeholders in `placeholder_names` are defined

        Removes any optional parts, whose placeholders are not in `placeholder_names`.
        """
        if len(self.optional_placeholders().difference(placeholder_names)) > 0:
            return []
        return self.sub_template.parts

    def remove_precursors(self, placeholders=None):
        return OptionalPart(self.sub_template.remove_precursors(placeholders))

    def __eq__(self, other):
        if not isinstance(other, OptionalPart):
            return NotImplemented
        return (self.sub_template == other.sub_template)


class TemplateParts:
    """
    The parts of a larger template
    """
    optional_re = re.compile(r'(\[.*?\])')
    requires_re = re.compile(r'(\{.*?\})')

    def __init__(self, parts: Sequence[Part]):
        if isinstance(parts, str):
            raise ValueError("Input to Template should be a sequence of parts; " +
                             "did you mean to call `TemplateParts.parse` instead?")
        self.parts = tuple(parts)

    @staticmethod
    @lru_cache(1000)
    def parse(text: str) -> "TemplateParts":
        """Parses a template string into its constituent parts

        Raises:
            ValueError: raised if a parsing error is

        Returns:
            TemplateParts: object that contains the parts of the template
        """

        parts = []
        for optional_parts in TemplateParts.optional_re.split(text):
            if len(optional_parts) > 0 and optional_parts[0] == '[' and optional_parts[-1] == ']':
                if '[' in optional_parts[1:-1] or ']' in optional_parts[1:-1]:
                    raise ValueError(f'Can not parse {text}, because unmatching square brackets were found')
                parts.append(OptionalPart(TemplateParts.parse(optional_parts[1:-1])))
            else:
                for required_parts in TemplateParts.requires_re.split(optional_parts):
                    if len(required_parts) > 0 and required_parts[0] == '{' and required_parts[-1] == '}':
                        if ':' in required_parts:
                            var_name, var_type = required_parts[1:-1].split(':')
                        else:
                            var_name, var_type = required_parts[1:-1], ''
                        parts.append(Required(var_name, var_type))
                    else:
                        parts.append(Literal(required_parts))
        return TemplateParts(parts)

    def __str__(self):
        """
        Returns the template as a string
        """
        return os.path.normpath(''.join([str(p) for p in self.parts]))

    def optional_placeholders(self, ) -> Set[str]:
        """Set of optional placeholders
        """
        if len(self.parts) == 0:
            return set()
        optionals = set.union(*[p.optional_placeholders() for p in self.parts])
        return optionals.difference(self.required_placeholders())

    def required_placeholders(self, ) -> Set[str]:
        """Set of required placeholders
        """
        if len(self.parts) == 0:
            return set()
        return set.union(*[p.required_placeholders() for p in self.parts])

    def ordered_placeholders(self, valid=None) -> Tuple[str]:
        """Sequence of all placeholders in order (can contain duplicates)
        """
        ordered_vars = []
        for p in self.parts:
            p.append_placeholders(ordered_vars, valid=valid)
        return ordered_vars

    def fill_known(self, placeholders: Placeholders, ignore_type=False) -> MyDataArray:
        """Fill in the known placeholders

        Any optional parts, where all placeholders have been filled will be automatically replaced
        """
        single, multi = placeholders.split()
        return self.remove_precursors(placeholders)._fill_known_helper(single, multi, ignore_type=ignore_type)

    def _fill_known_helper(self, single: Placeholders, multi: Placeholders, ignore_type=False) -> MyDataArray:
        new_template = self.fill_single_placeholders(single, ignore_type=ignore_type)
        for name in new_template.ordered_placeholders():
            use_name = multi.find_key(name)
            if use_name is None:
                continue
            new_multi = multi.copy()
            if use_name in multi.linkages:
                values = multi[multi.linkages[use_name]]
                keys = tuple(sorted(values.keys()))
                index = (keys, zip(*[values[k] for k in keys]))
                del new_multi[new_multi.linkages[use_name]]
            else:
                values = {use_name: list(multi[name])}
                index = (use_name, values[use_name])
                del new_multi[use_name]
            assert use_name is not None

            parts = []
            new_single = single.copy()
            for idx in range(len(values[use_name])):
                new_vals = {n: v[idx] for n, v in values.items()}
                new_single.mapping.update(new_vals)
                parts.append(new_template._fill_known_helper(new_single, new_multi, ignore_type=ignore_type))

            return MyDataArray.concat(parts, index)
        return MyDataArray(np.array(new_template), [])

    def fill_single_placeholders(self, placeholders: Placeholders, ignore_type=False) -> "TemplateParts":
        """
        Fills in placeholders with singular values

        Assumes that all placeholders are in fact singular
        """
        res = [p.fill_single_placeholders(placeholders, ignore_type=ignore_type) for p in self.parts]
        return TemplateParts(list(chain(*res)))

    def remove_optionals(self, optionals=None) -> "TemplateParts":
        """
        Removes any optionals containing the provided placeholders (default: remove all)
        """
        return TemplateParts([p for p in self.parts if not p.contains_optionals(optionals)])

    def all_matches(self, ) -> List[Dict[str, Any]]:
        """Finds all potential matches to existing templates

        Returns a list with the possible combination of values for the placeholders.
        """
        required = self.required_placeholders()
        optional = self.optional_placeholders()
        matches = []
        already_globbed = {}
        for defined_optionals in [c for n in range(len(optional) + 1) for c in combinations(optional, n)]:
            new_glob = str(self.fill_single_placeholders({**{req: '*' for req in required}, **{opt: '*' for opt in defined_optionals}}, ignore_type=True).remove_optionals())
            while '**' in new_glob:
                new_glob = new_glob.replace('**', '*')
            if new_glob not in already_globbed:
                already_globbed[new_glob] = glob(new_glob)
            res = []
            vars = required.union(defined_optionals)
            for p in self.parts:
                res.extend(p.for_defined(vars))
            parser = TemplateParts(res).get_parser()
            for fn in already_globbed[new_glob]:
                try:
                    placeholders = parser(fn)
                except ValueError:
                    continue
                for var_name in optional:
                    if var_name not in placeholders:
                        placeholders[var_name] = None
                matches.append(placeholders)
        return matches

    def resolve(self, placeholders, ignore_type=False) -> MyDataArray:
        """
        Resolves the template given a set of placeholders

        :param placeholders: mapping of placeholder names to values
        :param ignore_type: if True, ignore the type formatting when filling in placeholders
        :return: cleaned string
        """
        return self.fill_known(placeholders, ignore_type=ignore_type).map(lambda t: t.remove_optionals())

    def optional_subsets(self, ) -> Iterator["TemplateParts"]:
        """
        Yields template sub-sets with every combination optional placeholders
        """
        optionals = self.optional_placeholders()
        for n_optional in range(len(optionals) + 1):
            for exclude_optional in itertools.combinations(optionals, n_optional):
                yield self.remove_optionals(exclude_optional)

    def extract_placeholders(self, filename, known_vars=None):
        """
        Extracts the placeholder values from the filename

        :param filename: filename
        :param known_vars: already known placeholders
        :return: dictionary from placeholder names to string representations (unused placeholders set to None)
        """
        if known_vars is not None:
            template = self.fill_known(known_vars)
        else:
            template = self
        while '//' in filename:
            filename = filename.replace('//', '/')

        required = template.required_placeholders()
        optional = template.optional_placeholders()
        results = []
        for to_fill in template.optional_subsets():
            sub_re = str(to_fill.fill_known(
                {var: r'(\S+)' for var in required.union(optional)},
            ))
            while '//' in sub_re:
                sub_re = sub_re.replace('//', '/')
            sub_re = sub_re.replace('.', r'\.')
            match = re.match(sub_re, filename)
            if match is None:
                continue

            extracted_value = {}
            ordered_vars = to_fill.ordered_placeholders()
            assert len(ordered_vars) == len(match.groups())

            failed = False
            for var, value in zip(ordered_vars, match.groups()):
                if var in extracted_value:
                    if value != extracted_value[var]:
                        failed = True
                        break
                else:
                    extracted_value[var] = value
            if failed or any('/' in value for value in extracted_value.values()):
                continue
            for name in template.optional_placeholders():
                if name not in extracted_value:
                    extracted_value[name] = None
            if known_vars is not None:
                extracted_value.update(known_vars)
            results.append(extracted_value)
        if len(results) == 0:
            raise ValueError("{} did not match {}".format(filename, template))

        def score(placeholders):
            """
            The highest score is given to the set of placeholders that:

            1. has used the largest amount of optional placeholders
            2. has the shortest text within the placeholders (only used if equal at 1
            """
            number_used = len([v for v in placeholders.values() if v is not None])
            length_hint = sum([len(v) for v in placeholders.values() if v is not None])
            return number_used * 1000 - length_hint

        best = max(results, key=score)
        for var in results:
            if best != var and score(best) == score(var):
                raise KeyError("Multiple equivalent ways found to parse {} using {}".format(filename, template))
        return best

    def get_parser(self):
        if any(isinstance(p, OptionalPart) for p in self.parts):
            raise ValueError("Can not parse filename when there are optional parts in the template")
        mapping = {old_key: ''.join(new_key) for old_key, new_key in zip(self.required_placeholders(), itertools.product(*[string.ascii_letters] * 3))}
        reverse = {new_key: old_key for old_key, new_key in mapping.items()}
        cleaned = TemplateParts([Required(mapping[p.var_name], p.var_formatting) if isinstance(p, Required) else p for p in self.parts])
        parser = compile(str(cleaned), case_sensitive=True)

        def parse_filename(filename):
            result = parser.parse(filename)
            if result is None:
                raise ValueError(f"template string ({str(self)}) does not mach filename ({filename})")
            named = result.named
            if any(isinstance(value, str) and '/' in value for value in named.values()):
                raise ValueError("Placeholder can not span directories")
            return {reverse[key]: value for key, value in named.items()}
        
        return parse_filename

    def remove_precursors(self, placeholders=None):
        """Replaces keys to those existing in the placeholders

        If no placeholders provided all precursors are removed
        """
        return TemplateParts([p.remove_precursors(placeholders) for p in self.parts])

    def __eq__(self, other):
        if not isinstance(other, TemplateParts):
            return NotImplemented
        return (len(self.parts) == len(other.parts)) and all(p1 == p2 for p1, p2 in zip(self.parts, other.parts))


"""Parses a string representation of a FileTree
"""
import os.path as op
from .template import Template
from . import FileTree
from contextlib import contextmanager
from typing import Tuple, List, Optional
import re
import pkg_resources
from glob import glob


# searching for a file-tree

"""List of directories to look for FileTrees
"""
tree_directories = ['.']

available_subtrees = {}

scanned_plugins = set()


@contextmanager
def extra_tree_dirs(extra_dirs):
    """Temporarily insert ``extra_dirs`` to the beginning of :attr:`tree_directories`.

    :arg extra_dirs: Sequence of additional tree file directories to search.
    """

    global tree_directories

    old_tree_directories = list(tree_directories)

    tree_directories = list(extra_dirs) + list(tree_directories)

    try:
        yield
    finally:
        tree_directories = old_tree_directories


def scan_plugins():
    """Scan plugins for filetrees
    """
    for ep in pkg_resources.iter_entry_points("file_tree.trees"):
        if ep.module_name in scanned_plugins:
            continue
        plugin = ep.load()
        plugin()
        scanned_plugins.add(ep.module_name)


def search_tree(name: str) -> str:
    """
    Searches for the file defining the specific tree

    Iteratively searches through the directories in ``tree_directories`` till a file named ${name}.tree is found.
    If not found in ``tree_directories`` the filetrees in installed python packages will be searched.

    :param name: Name of the tree
    :return: string containing the filetree definition
    """
    for directory in tree_directories:
        filename = op.join(directory, name)
        for ext in ('.tree', ''):
            if op.exists(filename + ext):
                with open(filename + ext, 'r') as f:
                    return f.read()

    scan_plugins()

    for ext in ('.tree', ''):
        if name + ext in available_subtrees:
            return available_subtrees[name + ext]

    raise ValueError("No file tree found for %s" % name)


def list_all_trees() -> List[str]:
    """Generates a list of available sub-trees

    Lists trees available in ``tree_directories`` (default just the current directory) and in installed file-tree plugins (e.g., `file-tree-fsl`).
    """
    scan_plugins()
    trees = list(available_subtrees.keys())
    for directory in tree_directories:
        trees.extend(op.basename(fn) for fn in glob(op.join(directory, "*.tree")))
    return trees


# reading the file-tree

def read_file_tree_text(lines: List[str], top_level: Template) -> FileTree:
    """Parses the provided lines to read a FileTree

    See :func:`add_line_to_tree` for how individual lines are parsed

    Args:
        lines (List[str]): Individual lines read from a file-tree file
        top_level (Template): top-level template

    Returns:
        FileTree: tree read from the file
    """
    tree = FileTree.empty(top_level)
    current = []
    for line in lines:
        current = add_line_to_tree(tree, line, current)
    return tree


def add_line_to_tree(tree: FileTree, line: str, current: List[Tuple[int, Optional[str]]]) -> List[Tuple[int, Optional[str]]]:
    """Adds template or sub-tree defined on this file

    There are 5 options for the line:

        1. Empty lines or those containing only comments (start with #) do nothing.
        2. Templates have the form "  <unique part> (<short name>)" and are added as a new template (note that the <short name> is optional).
        3. Placeholder values have the form "<key> = <value>" and are stored as placeholder values in the tree.
        4. Sub-trees have the form "  -><tree name> [<placeholder>=<value>,...] (<short name>)" and are added as a new sub-tree
        5. Linkages between placeholder values are indicated by "&LINK <placeholder>,..."

    The parent directory of the new template or sub-tree is determined by the amount of white space.

    Args:
        tree (FileTree): tree containing all templates/sub-trees read so far (will be updated in place with new template or sub-tree)
        line (str): current line from the file-tree definition
        current (List[Tuple[int, Optional[str]]]): sequence of the possible parent directories and their indentation

    Returns:
        List[Tuple[int, Optional[str]]]: new sequence of the possible parent directories after reading the line
    """
    stripped = line.split('#')[0].strip()
    if len(stripped) == 0:
        return current
    nspaces = line.index(stripped)
    parent = get_parent(nspaces, current)
    new_current = [(n, name) for n, name in current if n < nspaces]

    if stripped[:2] == '->': 
        sub_tree, short_names = read_subtree_line(stripped)
        tree.add_subtree(sub_tree, short_names, parent)
        new_current.append((nspaces, None))
    elif '=' in stripped:
        key, value = [s.strip() for s in stripped.split('=')]
        if ',' in value:
            value = [None if v.strip() == "None" else v.strip() for v in value.split(',')]
        tree.update(inplace=True, **{key: value})
    elif stripped.startswith("&LINK"):
        keys = [k.strip() for k in stripped[5:].split(",")]
        tree.placeholders.link(*keys)
    else:
        if stripped[0] == '!':
            if nspaces != 0:
                raise ValueError(f"Defining a new top-level with '!' is only available at the top-level, but the line '{stripped}' is indented")
            stripped = stripped[1:]
            parent = None
        unique_part, short_names = read_line(stripped)
        if short_names is not None and len(short_names) == 0:
            raise ValueError("Empty short names are not allowed (they are used internally for the top-level directory")
        short_name = tree.add_template(unique_part, short_names, parent)
        new_current.append((nspaces, short_name))
    return new_current


def get_parent(nspaces: int, current: List[Tuple[int, str]]) -> str:
    """Determines the parent template based on the amount of whitespace

    Args:
        nspaces (int): amount of whitespace before the new line
        current (List[Tuple[int, str]]): sequence of possible parent directories and their indentation

    Raises:
        ValueError: raised of parent is a sub-tree rather than a template
        ValueError: raise if number of spaces does not match any existing directory

    Returns:
        str: empty string if the parent is the top-level directory; template short name otherwise
    """
    if len(current) == 0:
        return ""
    nspaces_max = current[-1][0]
    if nspaces > nspaces_max:
        if current[-1][1] is None:
            raise ValueError("Current line seems to be the child of a sub-tree, which is not supported.")
        return current[-1][1]

    for idx, (nspaces_template, _) in enumerate(current):
        if nspaces_template == nspaces:
            if idx == 0:
                return ""
            else:
                return current[idx - 1][1]
    raise ValueError("Number of spaces of current line does not match any previous lines.")


def check_forbidden_characters(text, characters, text_type):
    """
    Checks the text for forbidden characters

    raises ValueError if one is found

    :param text: string with the text
    :param characters: sequence of forbidden characters
    :param text_type: type of the text to raise in error message
    """
    bad = [character for character in characters if character in text]
    if len(bad) > 0:
        raise ValueError('Invalid character(s) "{}" in {}: {}'.format("".join(bad), text_type, text))


def read_line(line: str) -> Tuple[str, Optional[List[str]]]:
    """
    Parses line from the tree file

    :param line: input line from a ``*.tree`` file
    :return: Tuple with:

        - unique part of the filename
        - short name of the file (None if not provided)
    """
    if line.strip()[:1] == '->':
        return read_subtree_line(line)
    match = re.match(r'^(\s*)(\S*)\s*\((\S*)\)\s*$', line)
    if match is not None:
        gr = match.groups()
        check_forbidden_characters(gr[1], r'<>"|?*', 'file or directory name')
        if ',' in gr[2]:
            short_names = [name.strip() for name in gr[2].split(',') if len(name.strip()) > 0]
        else:
            short_names = [gr[2].strip()]
        return gr[1], short_names
    match = re.match(r'^(\s*)(\S*)\s*$', line)
    if match is not None:
        gr = match.groups()
        check_forbidden_characters(gr[1], r'<>"|?*', 'file or directory name')
        return gr[1], None
    raise ValueError('Unrecognized line %s' % line)


def read_subtree_line(line: str) -> Tuple[FileTree, Optional[List[str]]]:
    """
    Parses the line defining a sub_tree

    :param line: input line from a ``*.tree`` file
    :param template: containing template
    :return: Tuple with

        - sub_tree
        - short name of the sub_tree (None if not provided)
    """
    match = re.match(r'^(\s*)->\s*(\S*)(.*)\((\S*)\)', line)
    if match is None:
        match = re.match(r'^(\s*)->\s*(\S*)(.*)', line)
        if match is None:
            raise ValueError("Sub-tree line could not be parsed: {}".format(line.strip()))
        _, type_name, variables_str = match.groups()
        short_names = [None]
    else:
        _, type_name, variables_str, full_short_name = match.groups()
        check_forbidden_characters(full_short_name, r'(){}/', 'sub-tree name')
        if ',' in full_short_name:
            short_names = [name.strip() for name in full_short_name.split(',') if len(name.strip()) > 0]
        else:
            short_names = [full_short_name]

    check_forbidden_characters(type_name, r'<>:"/\|?*', 'filename of sub-tree')

    variables = {}
    if len(variables_str.strip()) != 0:
        for single_variable in variables_str.split(','):
            key, value = single_variable.split('=')
            variables[key.strip()] = value.strip()

    sub_tree = FileTree.read(type_name, **variables)
    return sub_tree, short_names



"""Defines the main FileTree object, which will be the main point of interaction
"""
from collections import defaultdict
import warnings
import xarray
from .template import Template, Placeholders, is_singular
from typing import Generator, Optional, Dict, Any, Sequence, Union
from pathlib import Path
import numpy as np
from functools import cmp_to_key
import string
from shutil import copyfile
import os
from warnings import warn


class FileTree:
    """Represents a structured directory

    The many methods can be split into 4 categories

        1. The template interface. Each path (file or directory) is reprsented by a :class:`Template <file_tree.template.Template>`, 
           which defines the filename with any unknown parts (e.g., subject ID) marked by placeholders. 
           Templates are accessed based on their key.

            - :meth:`get_template`: used to access a template based on its key.
            - :meth:`template_keys`: used to list all the template keys.
            - :meth:`add_template`: used to add a new template or overwrite an existing one.
            - :meth:`add_subtree`: can be used to add all the templates from a different tree to this one.
        
        2. The placeholder interface. Placeholders represent values to be filled into the placeholders.
           Each placeholder can be either undefined, have a singular value, or have a sequence of possible values.

            - You can access the :class:`placeholders dictionary-like object <file_tree.template.Placeholders>` directly through `FileTree.placeholders`
            - :meth:`update`: returns a new FileTree with updated placeholders or updates the placeholders in the current one.
            - :meth:`update_glob`: sets the placeholder values based on which files/directories exist on disk.
            - :meth:`iter_vars`: iterate over all possible values for the selected placeholders.
            - :meth:`iter`: iterate over all possible values for the placeholders that are part of a given template.
        
        3. Getting the actual filenames based on filling the placeholder values into the templates.

            - :meth:`get`: Returns a valid path by filling in all the placeholders in a template. 
              For this to work all placeholder values should be defined and singular.
            - :meth:`get_mult`: Returns array of all possible valid paths by filling in the placeholders in a template.
              Placeholder values can be singular or a sequence of possible values.
            - :meth:`get_mult_glob`: Returns array with existing paths on disk. 
              Placeholder values can be singular, a sequence of possible values, or undefined.
              In the latter case possible values for that placeholder are determined by checking the disk.
            - :meth:`fill`: Returns new FileTree with any singular values filled into the templates and removed from the placeholder dict.

        4. Input/output

            - :meth:`empty`: creates empty FileTree with no templates or placeholder values.
            - :meth:`read`: reads a new FileTree from a file.
            - :meth:`from_string`: reads a new FileTree from a string.
            - :meth:`write`: writes a FileTree to a file.
            - :meth:`to_string`: writes a FileTree to a string.
    """

    def __init__(self, templates: Dict[str, Template], placeholders: Dict[str, Any], return_path=False):
        self._templates = templates
        self.placeholders = Placeholders(placeholders)
        self.return_path = return_path

    # create new FileTree objects
    @classmethod
    def empty(cls, top_level: Union[str, Template]='.', return_path=False) -> "FileTree":
        """Creates a new empty FileTree containing only a top-level directory

        Args:
            top_level (str, Template, optional): Top-level directory that other templates will use as a reference. Defaults to current directory.

        Returns:
            FileTree: empty FileTree
        """
        if not isinstance(top_level, Template):
            top_level = Template(None, top_level)
        return cls({"": top_level}, {}, return_path=return_path)

    @classmethod
    def read(cls, name: str, top_level: Union[str, Template]='.', return_path=False, **placeholders) -> "FileTree":
        """Reads a filetree based on the given name

        Args:
            name (str): name of the filetree. Interpreted as:

                - a filename containing the tree definition if "name" or "name.tree" exist on disk
                - one of the trees in `tree_directories` if one of those contains "name" or "name.tree"
                - one of the tree in the plugin FileTree modules

            top_level (str): top-level directory name. Defaults to current directory. Set to parent template for sub-trees.
            placeholders (str->Any): maps placeholder names to their values

        Returns:
            FileTree: tree matching the definition in the file
        """
        from . import parse_tree
        if 'directory' in placeholders:
            warnings.warn(
                f"Setting the 'directory' placeholder to {placeholders['directory']}. " +
                "This differs from the behaviour of the old filetree in fslpy, which used the `directory` keyword to set the top-level directory. " +
                "If you want to do that, please use the new `top_level` keyword instead of `directory`."
                )
        for filename in (name, str(name) + '.tree'):
            if Path(filename).is_file():
                with open(filename, 'r') as f:
                    text = f.read()
                with parse_tree.extra_tree_dirs([Path(filename).parent]):
                    return cls.from_string(text, top_level, return_path=return_path, **placeholders)
        text = parse_tree.search_tree(name)
        return cls.from_string(text, top_level, return_path=return_path, **placeholders)

    @classmethod
    def from_string(cls, definition: str, top_level: Union[str, Template]='.', return_path=False, **placeholders) -> "FileTree":
        """Creates a FileTree based on the given definition

        Args:
            definition (str): A FileTree definition describing a structured directory
            top_level (str): top-level directory name. Defaults to current directory. Set to parent template for sub-trees.

        Returns:
            FileTree: tree matching the definition in the file
        """
        from . import parse_tree
        res = parse_tree.read_file_tree_text(definition.splitlines(), top_level).update(inplace=True, **placeholders)
        res.return_path = return_path
        return res

    def copy(self, ) -> "FileTree":
        """Creates a copy of the tree

        The dictionaries (templates, placeholders) are copied, but the values within them are not.

        Returns:
            FileTree: new tree object with identical templates, sub-trees and placeholders
        """
        new_tree = type(self)(dict(self._templates), Placeholders(self.placeholders), self.return_path)
        return new_tree

    # template interface
    def get_template(self, key: str) -> Template:
        """Returns the template corresponding to `key`.

        Raises:
            KeyError: if no template with that identifier is available

        Args:
            key (str): key identifying the template.

        Returns:
            Template: description of pathname with placeholders not filled in
        """
        return self._templates[key]

    @property
    def top_level(self, ):
        as_string = self.get_template("").unique_part
        if self.return_path:
            return Path(as_string)
        return str(as_string)

    @top_level.setter
    def top_level(self, value: str):
        self.get_template("").unique_part = str(value)

    def add_template(self, template_path: str, key: Optional[Union[str, Sequence[str]]]=None, parent: str="", overwrite=False) -> str:
        """Updates the FileTree with the new template

        Args:
            template_path: path name with respect to the parent (or top-level if no parent provided)
            key: key(s) to access this template in the future. Defaults to result from :meth:`Template.guess_key <file_tree.template.Template.guess_key>` 
                (i.e., the path basename without the extension).
            parent: if defined, `template_path` will be interpreted as relative to this template. 
                By default the top-level template is used as reference.
                To create a template unaffiliated with the rest of the tree, set `parent` to None. 
                Such a template should be an absolute path or relative to the current directory and can be used as parent for other templates
            overwrite: if True, overwrites any existing template rather than raising a ValueError. Defaults to False.

        Returns:
            str: one of the short names under which the template has been stored
        """
        parent_template = None if parent is None else self.get_template(parent)
        new_template = Template(parent_template, template_path)
        if parent_template is not None and new_template.as_path == parent_template.as_path:
            new_template = parent_template
        return self._add_actual_template(new_template, key, overwrite=overwrite)

    def _add_actual_template(self, template: Template, keys: Optional[Union[str, Sequence[str]]]=None, overwrite=False):
        if keys is None:
            keys = template.guess_key()
        if isinstance(keys, Path):
            keys = str(keys)
        if isinstance(keys, str):
            keys = [keys]
        for key in keys:
            if key in self._templates:
                if not overwrite:
                    raise ValueError(f"{self._templates[key]} already exists with short name '{key}', so cannot add {template}")
                old_template = self.get_template(key)
                for potential_child in self._templates.values():
                    if potential_child.parent is old_template:
                        potential_child.parent = template
            self._templates[key] = template
        return keys[0]

    def template_keys(self, only_leaves=False):
        """Returns the keys of all the templates in the FileTree

        Each key will be returned for templates with multiple keys.

        Args
            only_leaves (bool, optional): set to True to only return templates that do not have any children
        """
        if not only_leaves:
            return self._templates.keys()
        parents = {t.parent for t in self._templates.values() if t.parent is not None}
        return {key for key, template in self._templates.items() if template not in parents}

    def add_subtree(self, sub_tree: "FileTree", precursor: Optional[Union[str, Sequence[str]]]=(None, ), parent: Optional[str]="") -> None:
        """Updates the templates and the placeholders in place with those in sub_tree

        The top-level directory of the sub-tree will be replaced by the `parent` (unless set to None).
        The sub-tree templates will be available with the key "<precursor>/<original_key>",
        unless the precursor is None in which case they will be unchanged (which can easily lead to errors due to naming conflicts).

        What happens with the placeholder values of the sub-tree depends on whether the precursor is None or not:

            - if the precursor is None, any singular values are directly filled into the sub-tree templates. 
              Any placeholders with multiple values will be added to the top-level variable list (error is raised in case of conflicts).
            - if the precursor is a string, the templates are updated to look for "<precursor>/<original_placeholder>" and 
              all sub-tree placeholder values are also prepended with this precursor.
              Any template values with "<precursor>/<key>" will first look for that full key, but if that is undefined 
              they will fall back to "<key>" (see :class:`Placeholders <file_tree.template.Placeholders>`).
        
        The net effect of either of these procedures is that the sub-tree placeholder values will be used in that sub-tree,
        but will not affect templates defined elsewhere in the parent tree.
        If a placeholder is undefined in a sub-tree, it will be taken from the parent placeholder values (if available).

        Args:
            sub_tree (FileTree): tree to be added to the current one
            precursor (list(str or None)): name(s) of the sub-tree. Defaults to just adding the sub-tree to the main tree without precursor
            parent (str): key of the template used as top-level directory for the sub tree. 
                Defaults to top-level directory of the main tree. 
                Can be set to None for an independent tree.
        """
        if isinstance(precursor, str) or precursor is None:
            precursor = [precursor]
        for name in precursor:
            if name is None:
                add_string = ""
                sub_tree_fill = sub_tree.fill()
            else:
                add_string = name + "/"
                sub_tree_fill = sub_tree

            to_assign = dict(sub_tree_fill._templates)
            assigned = {"": (
                Template(None, to_assign[""].unique_part) 
                if parent is None else 
                self.get_template(parent)
            )}
            been_assigned = {to_assign[""]: ""}
            del to_assign[""]
            while len(to_assign) > 0:
                for key, old_template in list(to_assign.items()):
                    if old_template.parent is None:
                        parent_template = None
                    elif old_template.parent in been_assigned:
                        parent_key = been_assigned[old_template.parent]
                        parent_template = assigned[parent_key]
                    else:
                        continue
                    assigned[key] = Template(parent_template, old_template.unique_part).add_precursor(add_string)
                    been_assigned[old_template] = key
                    del to_assign[key]

            for key, template in assigned.items():
                if key == '' and name is None:
                    if parent is None:
                        for letter in string.ascii_letters:
                            label = f"tree_top_{letter}"
                            if (
                                label not in sub_tree_fill.template_keys() and
                                label not in self.template_keys()
                            ):
                                self._add_actual_template(template, label)
                                break
                    continue
                self._add_actual_template(template, add_string + key)

            if name is None:
                conflict = {key for key in sub_tree_fill.placeholders.keys() if key in self.placeholders}
                if len(conflict) > 0:
                    raise ValueError(f"Sub-tree placeholder values for {conflict} conflict with those set in the parent tree.")
            for old_key, old_value in sub_tree_fill.placeholders.items():
                if isinstance(old_key, str):
                    self.placeholders[add_string + old_key] = old_value
                else:
                    self.placeholders[frozenset(add_string + k for k in old_key)] = {add_string + k: v for k, v in old_value.items()}

    # placeholders interface
    def update(self, inplace=False, **placeholders) -> "FileTree":
        """Updates the placeholder values to be filled into the templates

        Args:
            inplace (bool): if True change the placeholders in-place (and return the FileTree itself); 
                by default a new FileTree is returned with the updated values without altering this one.
            **placeholders (Dict[str, Any]): maps placeholder names to their new values (None to mark placeholder as undefined)

        Returns:
            FileTree: Tree with updated placeholders (same tree as the current one if inplace is True)
        """ 
        new_tree = self if inplace else self.copy()
        new_tree.placeholders.update(placeholders)
        return new_tree

    def update_glob(self, template_key: Union[str, Sequence[str]], inplace=False) -> "FileTree":
        """Updates any undefined placeholders based on which files exist on disk for template

        Args:
            template_key (str or sequence of str): key(s) of the template(s) to use
            inplace (bool): if True change the placeholders in-place (and return the FileTree itself); 
                by default a new FileTree is returned with the updated values without altering this one.

        Returns:
            FileTree: Tree with updated placeholders (same tree as the current one if inplace is True)
        """
        if isinstance(template_key, str):
            template_key = [template_key]
        new_placeholders = defaultdict(set)
        for key in template_key:
            template = self.get_template(key)
            from_template = template.get_all_placeholders(self.placeholders)
            for name, values in from_template.items():
                new_placeholders[name] = new_placeholders[name].union(values)

        def cmp(item1, item2):
            if item1 is None:
                return -1
            if item2 is None:
                return 1
            if item1 < item2:
                return -1
            if item1 > item2:
                return 1
            return 0
        return self.update(inplace, **{k: sorted(v, key=cmp_to_key(cmp)) for k, v in new_placeholders.items()})

    # Extract paths
    def get(self, key: str, make_dir=False) -> Union[str, Path]:
        """Returns template with placeholder values filled in

        Args:
            key (str): identifier for the template
            make_dir (bool, optional): If set to True, create the parent directory of the returned path.

        Returns:
            Path: Filled in template as Path object. 
                Returned as a `pathlib.Path` object if `FileTree.return_path` is True. 
                Otherwise a string is returned.
        """
        path = self.get_template(key).format_single(self.placeholders)
        if make_dir:
            Path(path).parent.mkdir(parents=True, exist_ok=True)
        if self.return_path:
            return Path(path)
        return path

    def get_mult(self, key: Union[str, Sequence[str]], filter=False, make_dir=False) -> Union[xarray.DataArray, xarray.Dataset]:
        """Returns array of paths with all possible values filled in for the placeholders 

        Singular placeholder values are filled into the template directly. 
        For each placeholder with multiple values a dimension is added to the output array.
        This dimension will have the name of the placeholder and labels corresponding to the possible values (see http://xarray.pydata.org/en/stable/).
        The precense of required, undefined placeholders will lead to an error 
        (see :meth:`get_mult_glob` or :meth:`update_glob` to set these placeholders based on which files exist on disk).

        Args:
            key (str, Sequence[str]): identifier(s) for the template.
            filter (bool, optional): If Set to True, will filter out any non-existent files.
                If the return type is strings, non-existent entries will be empty strings.
                If the return type is Path objects, non-existent entries will be None.
                Note that the default behaviour is opposite from :meth:`get_mult_glob`.
            make_dir (bool, optional): If set to True, create the parent directory for each returned path.

        Returns:
            xarray.DataArray, xarray.Dataset: For a single key returns all possible paths in an xarray DataArray. 
                For multiple keys it returns the combination of them in an xarray Dataset.
                Each element of in the xarray is a `pathlib.Path` object if `FileTree.return_path` is True.
                Otherwise the xarray will contain the paths as strings.
        """
        if isinstance(key, str):
            paths = self.get_template(key).format_mult(self.placeholders, filter=filter)
            paths.name = key
            if make_dir:
                for path in paths.data.flat:
                    if path is not None:
                        Path(path).parent.mkdir(parents=True, exists_ok=True)
            if self.return_path:
                return xarray.apply_ufunc(lambda p: None if p == "" else Path(p), paths, vectorize=True)
            return paths
        else:
            return xarray.merge([self.get_mult(k, filter=filter, make_dir=make_dir) for k in key], join='exact')

    def get_mult_glob(self, key: Union[str, Sequence[str]]) -> Union[xarray.DataArray, xarray.Dataset]:
        """Returns array of paths with all possible values filled in for the placeholders 

        Singular placeholder values are filled into the template directly. 
        For each placeholder with multiple values a dimension is added to the output array.
        This dimension will have the name of the placeholder and labels corresponding to the possible values (see http://xarray.pydata.org/en/stable/).
        The possible values for undefined placeholders will be determined by which files actually exist on disk.

        The same result can be obtained by calling `self.update_glob(key).get_mult(key, filter=True)`.
        However calling this method is more efficient, because it only has to check the disk for which files exist once.

        Args:
            key (str, Sequence[str]): identifier(s) for the template.

        Returns:
            xarray.DataArray, xarray.Dataset: For a single key returns all possible paths in an xarray DataArray. 
                For multiple keys it returns the combination of them in an xarray Dataset.
                Each element of in the xarray is a `pathlib.Path` object if `FileTree.return_path` is True.
                Otherwise the xarray will contain the paths as strings.
        """
        if isinstance(key, str):
            template = self.get_template(key)
            matches = template.all_matches(self.placeholders)

            new_placeholders = Placeholders(self.placeholders)
            new_placeholders.update(template.get_all_placeholders(self.placeholders, matches=matches))

            paths = template.format_mult(new_placeholders, filter=True, matches=matches)
            paths.name = key
            if self.return_path:
                return paths
            res = xarray.apply_ufunc(lambda p: "" if p is None else str(p), paths, vectorize=True)
            return res
        else:
            return xarray.merge([self.get_mult_glob(k) for k in key], join='outer', fill_value=None if self.return_path else "")


    def fill(self, keep_optionals=True) -> "FileTree":
        """Fills in singular placeholder values.

        Args:
            keep_optionals: if True keep optional parameters that have not been set

        Returns:
            FileTree: new tree with singular placeholder values filled into the templates and removed from the placeholder dict
        """
        new_templates = {}
        to_assign = dict(self._templates)
        get_keys = {t: k for k, t in to_assign.items()}
        while len(to_assign) > 0:
            for key, old_template in list(to_assign.items()):
                if old_template.parent is None:
                    new_parent = None
                elif get_keys[old_template.parent] in new_templates:
                    new_parent = new_templates[get_keys[old_template.parent]]
                else:
                    continue
                new_templates[key] = Template(new_parent, str(Template(None, old_template.unique_part).format_single(self.placeholders, check=False, keep_optionals=keep_optionals)))
                del to_assign[key]
        _, mult_placeholders = self.placeholders.split()
        return FileTree(new_templates, mult_placeholders)

    # iteration
    def iter_vars(self, placeholders: Sequence[str]) -> Generator["FileTree", None, None]:
        """Iterate over the placeholder placeholder names

        Args:
            placeholders (Sequence[str]): sequence of placeholder names to iterate over

        Returns:
            Generator[FileTree]: yields trees, where each placeholder only has a single possible value
        """
        for placeholders in self.placeholders.iter_over(placeholders):
            yield FileTree(self._templates, placeholders, self.return_path)

    def iter(self, template: str, check_exists: bool=False) -> Generator["FileTree", None, None]:
        """Iterate over trees containng all possible values for template

        Args:
            template (str): short name identifier of the template
            check_exists (bool): set to True to only return trees for which the template actually exists
        
        Returns:
            Generator[FileTree]: yields trees, where each placeholder in given template only has a single possible value
        """
        placeholders = self.get_template(template).placeholders(valid=self.placeholders.keys())
        for tree in self.iter_vars(placeholders):
            if not check_exists or tree.get(template).exists:
                yield tree

    # convert to string
    def to_string(self, indentation=4) -> str:
        """Converts FileTree into a valid filetree definition

        An identical FileTree can be created by running :meth:`from_string` on the resulting string.

        Args:
            indentation (int, optional): Number of spaces to use for indendation. Defaults to 4.
        """
        placeholder_lines = []
        for key, value in self.placeholders.items():
            if value is None:
                continue
            if isinstance(key, frozenset):
                # linked placeholder
                for linked_key, linked_value in value.items():
                    placeholder_lines.append(f"{linked_key} = {', '.join([str(v) for v in linked_value])}")
                placeholder_lines.append(f"&LINK {', '.join(key)}")

            elif np.array(value).ndim == 1:
                placeholder_lines.append(f"{key} = {', '.join([str(v) for v in value])}")
            else:
                placeholder_lines.append(f"{key} = {value}")
        lines = ['\n'.join(placeholder_lines)]

        top_level = sorted([(key, template) for key, template in self._templates.items() if template.parent is None], key=lambda k: k[0])
        already_done = set()
        for _, t in top_level:
            if t not in already_done:
                lines.append(t.as_multi_line(self._templates, indentation=indentation))
                already_done.add(t)
        return '\n\n'.join(lines)

    def write(self, filename, indentation=4):
        """Writes the FileTree to a disk as a text file

        The first few lines will contain the placeholders.
        The remaining lines will contain the actual FileTree with all the templates (including sub-trees).
        The top-level directory is not stored in the file and hence will need to be provided when reading the tree from the file.

        Args:
            filename (str or Path): where to store the file (directory should exist already)
            indentation (int, optional): Number of spaces to use in indendation. Defaults to 4.
        """
        with open(filename, 'w') as f:
            f.write(self.to_string(indentation=indentation))


def convert(src_tree: FileTree, target_tree: Optional[FileTree]=None, keys=None, symlink=False, overwrite=False):
    """
    Copies or links files defined in `keys` from the `src_tree` to the `target_tree`.

    Given two example trees
    
        - source:: 

            subject = A,B

            sub-{subject}
                data
                    T1w.nii.gz
                    FLAIR.nii.gz

        - target:: 

            subject = A,B

            data
                sub-{subject}
                    {subject}-T1w.nii.gz (T1w)
                    {subject}-T2w.nii.gz (T2w)

    Given pre-existing data matching the source tree::

        .
        ├── sub-A
        │   └── data
        │       ├── FLAIR.nii.gz
        │       └── T1w.nii.gz
        └── sub-B
            └── data
                ├── FLAIR.nii.gz
                └── T1w.nii.gz

    We can do the following conversions:

        - `convert(source, target)`: copies all matching keys from `source` to `target`. 
            This will only copy the "T1w.nii.gz" files, because they are the only match in the template keys.
            Note that the `data` template key also matches between the two trees, but this template is not a leave, so is ignored.
        - `convert(source, target, keys=['T1w', ('FLAIR', 'T2w')]): copies the "T1w.nii.gz" files from `source` to `target` and 
            copies the "FLAIR.nii.gz" in `source` to "T2w..nii.gz" in `target.
        - `convert(source.update(subject='B'), source.update(subject='C')): creates a new "data/sub-C" directory and 
            copies all the data from "data/sub-B" into that directory.
        - `convert(source, keys=[('FLAIR', 'T1w')], overwrite=True)`: copies the "FLAIR.nii.gz" into the "T1w.nii.gz" files overwriting the originals.

    Warnings are raised in two cases:
    - if a source file is missing
    - if a target file already exists and `overwrite` is False

    Args:
        src_tree: prepopulated filetree with the source files
        target_tree: filetree that will be populated. Defaults to same as `src_tree`.
        keys (collection of str or (str, str), optional): collection of template keys to transfer from `src_tree` to `target_tree`. Defaults to all templates keys shared between `src_tree` and `target_tree`.
        symlink: if set to true links the files rather than copying them
        overwrite: if set to True overwrite any existing files
    """
    if target_tree is None and keys is None:
        raise ValueError("Conversion requires either `target_tree` or `keys` to be set")
    src_tree = src_tree.copy()
    if target_tree is None:
        target_tree = src_tree
    target_tree = target_tree.copy()
    if keys is None:
        keys = set(src_tree.template_keys(only_leaves=True)).intersection(target_tree.template_keys(only_leaves=True))

    full_keys = {(key_definition, key_definition) if isinstance(key_definition, str) else key_definition for key_definition in keys}
    for src_key, target_key in full_keys:
        # ensure template placeholders are consistent between source and target tree
        for placeholder in src_tree.get_template(src_key).placeholders() + target_tree.get_template(target_key).placeholders():
            if placeholder not in src_tree.placeholders:
                if placeholder not in target_tree.placeholders:
                    raise ValueError(f"Can not convert template {src_key}, because no values have been set for {placeholder}")
                src_tree.placeholders[placeholder] = target_tree.placeholders[placeholder]
            elif placeholder not in target_tree.placeholders:
                target_tree.placeholders[placeholder] = src_tree.placeholders[placeholder]
            nsrc = -1 if is_singular(src_tree.placeholders[placeholder]) else len(src_tree.placeholders[placeholder])
            ntarget = -1 if is_singular(target_tree.placeholders[placeholder]) else len(target_tree.placeholders[placeholder])
            if nsrc != ntarget:
                raise ValueError(f"Number of possible values for {placeholder} do not match between source and target tree")

        # ensure non-singular placeholders match
        src_non_singular = {p for p in src_tree.get_template(src_key).placeholders() if not is_singular(src_tree.placeholders[p])}
        target_non_singular = {p for p in target_tree.get_template(target_key).placeholders() if not is_singular(target_tree.placeholders[p])}

        diff = src_non_singular.difference(target_non_singular)
        if len(diff) > 0:
            raise ValueError("Placholders {diff} in source template {src_key} has no equivalent in target template {target_key}")
        diff = target_non_singular.difference(src_non_singular)
        if len(diff) > 0:
            raise ValueError("Placholders {diff} in target template {target_key} has no equivalent in source template {src_key}")

    # all checks have passed; let's get to work
    to_warn_about = ([], [])

    for src_key, target_key in sorted(full_keys):
        iter_placeholders = sorted({p for p in src_tree.get_template(src_key).placeholders() if not is_singular(src_tree.placeholders[p])})
        for single_src_tree, single_target_tree in zip(
            src_tree.iter_vars(iter_placeholders),
            target_tree.iter_vars(iter_placeholders),
        ):
            _convert_file(
                Path(single_src_tree.get(src_key)), 
                Path(single_target_tree.get(target_key)),
                to_warn_about,
                symlink=symlink,
                overwrite=overwrite,
            )
    if len(to_warn_about[0]) > 0:
        warn(f"Following source files were not found during FileTree conversion: {to_warn_about[0]}")
    if len(to_warn_about[1]) > 0:
        warn(f"Following target files already existed during FileTree conversion: {to_warn_about[1]}")


def _convert_file(source: Path, target: Path, to_warn_about, symlink=False, overwrite=False):
    """
    Copies or links `source` file to `target` file

    Helper function for :func:`convert`
    """
    if not source.exists():
        to_warn_about[0].append(source)
        return
    if target.exists():
        if not overwrite:
            to_warn_about[1].append(target)
            return
        os.remove(target)
    target.parent.mkdir(parents=True, exist_ok=True)
    if not symlink:
        copyfile(source, target, follow_symlinks=False)
    elif source.is_absolute():
        target.symlink_to(source)
    else:
        target.symlink_to(source.relative_to(target.parent))



import setuptools
import os.path as op


setuptools.setup(
    name="file-tree",
    version="0.4.1",
    url="https://git.fmrib.ox.ac.uk/ndcn0236/file-tree",

    author="Michiel Cottaar",
    author_email="MichielCottaar@protonmail.com",

    description="Describe structure directory for pipeline",
    long_description=open('README.md').read(),

    package_dir={'': "src"},

    packages=setuptools.find_packages("src", exclude=('tests*', '*.egg-info')),

    zip_safe=False,

    install_requires=[
        "numpy", "xarray", "pandas", "parse",
    ],

    project_urls = {
        "Documentation": "https://open.win.ox.ac.uk/pages/ndcn0236/file-tree/",
    },

    include_package_data=True,

    license='MIT',
    license_file='LICENSE',

    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: MIT License',
    ],
)
